#pragma once

//**************************************************************************/
// DESCRIPTION: Includes for Plugins
// AUTHOR: 
//***************************************************************************/


#include "3dsmaxsdk_preinclude.h"
#include "Max.h"
#include "resource.h"
#include "istdplug.h"
#include "iparamb2.h"
#include "iparamm2.h"
#include "INodeTransformMonitor.h"
#include "expr.h"
#include "notify.h"

#define IS_SUPMAX2012 (MAX_RELEASE>=14000)
#define IS_SUPMAX2013 (MAX_RELEASE>=15000)
#define IS_SUPMAX2015 (MAX_RELEASE>=17000)

#if IS_SUPMAX2012
	#include "maxscript\maxscript.h" 
#else
	#include "maxscrpt\maxscrpt.h" 
#endif

#if IS_SUPMAX2013
	#define end p_end	
#endif


extern TCHAR *GetString(int id);

extern HINSTANCE hInstance;

#define IKFKSolver_CLASS_ID		Class_ID(0x4ec52e6b, 0x7ed63cad)
#define FKCtrl_CLASS_ID			Class_ID(0xf266d92, 0x3c2f6126)
#define IKCtrl_CLASS_ID			Class_ID(0x521f77e0, 0x4f440f13)
#define AnimPivot_CLASS_ID		Class_ID(0x3d1aff84, 0x11124a76)
#define AnimPivotTarget_CLASS_ID	Class_ID(0x3d1aff85, 0x11124a77)	

//Solver refs
#define PBLOCK_REF		0
#define GOAL_REF		1
#define TARGET_REF		2
#define MAIN_REF		3
#define SEC_REF			4
#define LAST_REF		5
#define NONUSED_REF		6

//IK FK Ctrls Refs

#define LOCALPBLOCK_REF		0
#define LOCALCTRL_REF		1
#define SOLVERPBLOCK_REF	2
#define SOLVER_REF			3

//Extra refs
#define ANIMPIVOT_PIVOT_REF 0
#define ANIMPIVOT_TM_REF	1

//IK FK SUBANIMS (used for showing the keys of all the nodes in every node)
#define SHAREDSUB_REF		1

#define CTRL_NSUBS		3

#define SOLVER_NSUBS		1
#define SOLVER_ALL_NSUBS	2

#define NSOLVER_REFS	7
#define FK_REFS			4
#define IK_REFS			4

#define PICK_TARGET			0
#define PICK_GOALPARENT		1

#define GOAL_COLOR		RGB(99,150,224)
#define SOLVER_COLOR	RGB(255,255,255)

#define IK_COLOR		Point3(0.0,0.0,1.0)
#define FK_COLOR		Point3(1.0,0.0,0.0)

#define REFMSG_NODE_DELETED 0x18F91902
#define REFMSG_NODE_CHANGED 0x3071552C

#define FORCE_SOLVE (1<<0)
#define FULL_IKGOAL (1<<1)
#define	FULL_IKTARGET (1<<2)
#define	FULL_FKMAIN (1<<3)
#define	FULL_FKSEC (1<<4)

#define ANGLE_TRESHOLD 0.017f


#define IKFKSOLVER_INTERFACE Interface_ID(0x20e574a3, 0x6b485b6c)

enum { 
	IKFKSolver_params,
	IKFKSolver_options,
	IKFKSolver_display,
};
enum {
	pb_IKFKBlend,
	pb_IKFKSwivelAngle,
	pb_scaleMult,
	pb_IKStretch,
	pb_matchMainPreferred,
	pb_matchSecPreferred,
	pb_matchScaleMult,
	pb_parentSpace,
	pb_prefTreshold,
	pb_displayGoal,
	pb_displayTarget,
	pb_displayIKFK,
	pb_nGoalParent,
	pb_showAllKeys,
	pb_matchIKFK,
	pb_matchTargetSwivel,
};


	



// mixin interface 
class IKFKSolverInterface : public FPMixinInterface 
{ 
public:
	enum { 
		fnMatch, 
		fnSetFK,
		fnSetIK,
		fnGetMainFKNode,
		fnGetSecFKNode,
		fnGetLastFKNode,
		fnGetGoal,
		fnGetTarget,
		fnGetGoalParent,
		fnSetTarget, 
		fnSetGoalParent,
   end 
	}; 
   BEGIN_FUNCTION_MAP 
	 VFN_0(fnMatch, Match); 
	 VFN_0(fnSetFK, SetFK);
	 VFN_0(fnSetIK, SetIK);
	 FN_0(fnGetMainFKNode,TYPE_INODE, GetMainFKNode);
	 FN_0(fnGetSecFKNode,TYPE_INODE, GetSecFKNode);
	 FN_0(fnGetLastFKNode,TYPE_INODE, GetLastFKNode);
	 FN_0(fnGetGoal,TYPE_INODE, GetGoal);
	 FN_0(fnGetTarget,TYPE_INODE, GetTarget);
	 FN_0(fnGetGoalParent,TYPE_INODE, GetGoalParent); 
	 VFN_1(fnSetTarget, SetTargetNode,TYPE_INODE);
	 VFN_1(fnSetGoalParent, SetGoalParentNode,TYPE_INODE); 
   END_FUNCTION_MAP 

   virtual FPInterfaceDesc* GetDesc()=0;
   virtual void Match()=0; 
   virtual void SetFK()=0; 
   virtual void SetIK()=0; 
   virtual INode *GetMainFKNode ()=0;
   virtual INode *GetSecFKNode ()=0;
   virtual INode *GetLastFKNode ()=0;
   virtual INode *GetGoal ()=0;
   virtual INode *GetTarget ()=0;
   virtual INode *GetGoalParent ()=0;
   virtual void SetTargetNode(INode *theNode)=0; 
   virtual void SetGoalParentNode(INode *theNode)=0; 
};

class IKFKSolver : public ReferenceTarget {
	public:

		IObjParam *ip;
		HWND hwnd1,hwnd2;
		// Parameter block
		IParamBlock2			*pblock;	//ref 0
		ReferenceTarget			*goalNode;
		ReferenceTarget			*targetNode;
		ReferenceTarget			*mainNode;
		ReferenceTarget			*secNode;
		ReferenceTarget			*lastNode;
		Tab<ReferenceTarget*>	goalParentNode;
		

		BOOL canUpdate;
		BOOL canUpdateSwivel;
		BOOL notProcessing;//used to avoid dependency loops
		BOOL notNotifyingKeySel;//used to avoid dependency loops when notifyin key selection change

		int setTMCount; //used when several nodes of the solver are selected and transformed at the same time

		TimeValue prevt;
		TimeValue maint;
		TimeValue sect;
		TimeValue goalt;
		BOOL hasChanged;
		
		BOOL mustRemoveGoalParent;
		int nGoalParentToRemove;

		Tab<Animatable*>	allCtrls;

		// the Solved Values are stored in these variables for local usage
		Matrix3 mainTM, secTM, lastTM, goalTM, targetTM;
		Matrix3 *gParentTM;
		int solvedNumGoalParent;
		 //Used in the CalculateIK function to store rotation values of main and second fk nodes
		//used later in FKMatch
		Quat mainRot, secRot;

		// Loading/Saving
		IOResult Load(ILoad *iload) 
		{
			return IO_OK;
		}
		IOResult Save(ISave *isave) {return IO_OK;}

		//From Animatable
		Class_ID ClassID() {return IKFKSolver_CLASS_ID;}		
		SClass_ID SuperClassID() { return REF_TARGET_CLASS_ID; }
		void GetClassName(TSTR& s) {s = GetString(IDS_SOLVER_CLASS_NAME);}
		BOOL BypassTreeView() {return TRUE;}
		


		//to display controller params on the motion panel 
		void BeginEditParams( IObjParam *ip, ULONG flags,Animatable *prev); 
		// to undisplay 
		void EndEditParams( IObjParam *ip, ULONG flags,Animatable *next ); 
		

		RefTargetHandle Clone( RemapDir &remap );

#if IS_SUPMAX2015
		RefResult NotifyRefChanged(const Interval& changeInt, RefTargetHandle hTarget, 
			PartID& partID,  RefMessage message, BOOL propagate);
#else
		RefResult NotifyRefChanged(Interval changeInt, RefTargetHandle hTarget, 
			PartID& partID,  RefMessage message);
#endif

		int GoalParentCount()
		{
			int n=0;
			if (pblock)
				pblock->GetValue(pb_nGoalParent,0,n,FOREVER);
			return n;
		}
		int NumRefs() { if (pblock) return NSOLVER_REFS+goalParentNode.Count()/*GoalParentCount()*/; else return NSOLVER_REFS;}
		RefTargetHandle GetReference(int i);
		void SetReference(int i, RefTargetHandle rtarg);
		
		int NumSubs();
		Animatable* SubAnim(int i);
		TSTR SubAnimName(int i);
		int SubNumToRefNum(int subNum){return -1;}
		BOOL CanAssignController(int subAnim){return FALSE;}

		int	NumParamBlocks() { return 1; }					// return number of ParamBlocks in this instance
		IParamBlock2* GetParamBlock(int i) {if (i==0) return pblock; else return	NULL;} // return i'th ParamBlock
		// return id'd ParamBlock 
		IParamBlock2* GetParamBlockByID(BlockID id) {
			if (pblock && pblock->ID()==id)
				return pblock;
			return NULL;
		}

		int IsKeyable(){return TRUE;}
		
		//NodeMonitor Functions
		INode *GetMainNode();
		INode *GetSecNode();
		INode *GetLastNode();
		INode *GetGoalNode();
		INode *GetTargetNode();
		INode *GetGoalParentNode(int index);
		void SetMainNode(INode *theNode);
		void SetSecNode(INode *theNode);
		void SetLastNode(INode *theNode);
		void SetGoalNode(INode *theNode);
		void SetTargetNode(INode *theNode);
		void SetGoalParentNode(INode *theNode);
		void SetGoalParentNode(INode *theNode, int index);
		
		void SetGoalParentCount(int n);

		

		//Solver functions
		void CalculateIK(TimeValue t, Interval &valid, unsigned char solveFlags);
		BOOL Solve(TimeValue t,unsigned char solveFlags=0);

		//Match related functions
		void Match(TimeValue t);
		void FKMatch(TimeValue t,Control *theCtrl=NULL);
		void IKMatch(TimeValue t);
		void SwivelMatch(TimeValue t);
		void TargetMatch(TimeValue t);
		Point3 TargetMatch1(TimeValue t, Point3 prevLastPos);
		void TargetMatch2(TimeValue t, Point3 newTargetPos);
		void MainPrefMatch(TimeValue t);
		void SecPrefMatch(TimeValue t,BOOL forceMatch=FALSE);
		void ScaleMultMatch(TimeValue t,Control *theCtrl=NULL);
		int NumSelectedIKGoal();
		int NumSelectedIKTarget();
		int NumSelectedFK();

		void NotifyChange();
		void NotifySubAnimChange();
		void NotifyKeySelChange();
		Matrix3 GetGoalParentTM(TimeValue t);
		Matrix3 GetTargetParentTM(TimeValue t, INode *theNode, Interval &valid);
		Matrix3 GetParentTM(TimeValue t,INode *theNode,Interval &valid);	

		Matrix3 MakeLookAtMatrix(Point3 origin, Point3 target,Point3 lookup);
		Matrix3 MakeLookAtMatrix2(Point3 lookVector, Point3 upVector,Point3 position);
		Quat MakeLookAtQuat(Point3 origin, Point3 target,Matrix3 lookup);
		Quat RotVector(Point3 v1, Point3 v2);
		float GetCloserAngle(float prevAngle, float newAngle);
		void GetCtrlPRS(TimeValue t,Control *ctrl,Point3 &pos,Quat &rot,Point3 &scale, Interval &valid);
		Point3 GetTMScale(Matrix3 theTM);
		Quat MatrixToQuat(Matrix3 theTM);
		void CheckHierarchy();
		void RestoreCtrl(INode *theNode,Control *theCtrl);
		BOOL CanUpdate();
		void SetCanUpdate(BOOL canUpdate);
		BOOL isValidTarget(INode* theNode);
		BOOL isValidGoalParent(INode* theNode);
		BOOL DisplayIKFK(TimeValue t,INode* inode,ViewExp* vpt,int flags);

		BOOL FindCtrl (Animatable *ctrl, Tab<Animatable*> &theCtrls);
		void AppendKeyableCtrls (Animatable *ctrl, Tab<Animatable*> &theCtrls);
		void RefreshKeys();
		int  GetLocalKeyIndex(int index, int &ctrlIndex);
		//keys functions, this functions are used to have access to shared keys (when the show all keys option is enabled)
		int  NumKeys ();
		TimeValue  GetKeyTime (int index);
		int  GetKeyIndex (TimeValue t);
		BOOL  GetNextKeyTime (TimeValue t, DWORD flags, TimeValue &nt) ;
		void  CopyKeysFromTime (TimeValue src, TimeValue dst, DWORD flags);
		void  DeleteKeyAtTime (TimeValue t);
		BOOL  IsKeyAtTime (TimeValue t, DWORD flags);
		int  GetKeyTimes (Tab< TimeValue > &times, Interval range, DWORD flags);
		int  GetKeySelState (BitArray &sel, Interval range, DWORD flags);
		void  MapKeys (TimeMap *map, DWORD flags);
		void  DeleteKeys (DWORD flags);
		void  DeleteKeyByIndex (int index);
		void  SelectKeys (TrackHitTab &sel, DWORD flags);
		void  SelectKeyByIndex (int i, BOOL sel);
		BOOL  IsKeySelected (int i);
		void  FlagKey (TrackHitRecord hit);
		int  GetFlagKeyIndex ();
		int  NumSelKeys ();
		void  CloneSelectedKeys (BOOL offset=FALSE);
		void  AddNewKey (TimeValue t, DWORD flags);
		

		//Constructor/Destructor
		IKFKSolver(BOOL loading=FALSE);
		IKFKSolver(INode *mainBone,INode *secBone,INode *lastBone,INode *goal=NULL);
		~IKFKSolver();

		void DeleteThis() { delete this; }	

};



class IKFKSolverClassDesc : public ClassDesc2 
{
public:
	virtual int IsPublic() 							{ return FALSE; }
	virtual void* Create(BOOL loading = FALSE) 		{ return new IKFKSolver(loading); }
	virtual const TCHAR *	ClassName() 			{ return GetString(IDS_SOLVER_CLASS_NAME); }
	virtual SClass_ID SuperClassID() 				{ return REF_TARGET_CLASS_ID; }
	virtual Class_ID ClassID() 						{ return IKFKSolver_CLASS_ID; }
	virtual const TCHAR* Category() 				{ return GetString(IDS_CATEGORY); }

	virtual const TCHAR* InternalName() 			{ return _T("IKFKSolver"); }	// returns fixed parsable name (scripter-visible name)
	virtual HINSTANCE HInstance() 					{ return hInstance; }					// returns owning module handle
	

};

class IKFKSolverParamsDlgProc: public ParamMap2UserDlgProc
{
public:
	IKFKSolver *solver;
	IKFKSolverParamsDlgProc(IKFKSolver *thesolver){solver=thesolver;}
	virtual INT_PTR  DlgProc (TimeValue t, IParamMap2 *map, HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
	virtual void  DeleteThis (){delete this;}
};
class IKFKSolverOptionsDlgProc: public ParamMap2UserDlgProc
{
public:
	IKFKSolver *solver;
	IKFKSolverOptionsDlgProc(IKFKSolver *thesolver){solver=thesolver;}
	virtual INT_PTR  DlgProc (TimeValue t, IParamMap2 *map, HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
	virtual void  DeleteThis (){delete this;}
};
class IKFKSolverPickNode : public PickModeCallback, public PickNodeCallback 
{
public:			
	IKFKSolver *solver;
	DWORD	pickType;
	IKFKSolverPickNode(IKFKSolver *thesolver,DWORD type){solver=thesolver;pickType=type;}
	BOOL HitTest(IObjParam *ip, HWND hWnd, ViewExp *vpt, IPoint2 m, int flags);		
	BOOL Pick(IObjParam *ip,ViewExp *vpt);		
	void EnterMode(IObjParam *ip);
	void ExitMode(IObjParam *ip);		
	BOOL Filter(INode *node);
	PickNodeCallback *GetFilter() {return this;}
	BOOL RightClick(IObjParam *ip,ViewExp *vpt) {return TRUE;}
};
class IKFKSolverPostPatchProc : public PostPatchProc
{
public:
	IKFKSolver *solver;
	IKFKSolverPostPatchProc(IKFKSolver *theSolver){solver=theSolver;}
	int  Proc (RemapDir &remap);
};

class IKFKSolverClassAccessor : public PBAccessor
{
public:
	void Set(PB2Value &v, ReferenceMaker *owner, ParamID id, int tabIndex, TimeValue t);
};

class FKCtrlClassAccessor : public PBAccessor
{
public:
	void Get (PB2Value &v, ReferenceMaker *owner, ParamID id, int tabIndex, TimeValue t, Interval &valid);
	void Set(PB2Value &v, ReferenceMaker *owner, ParamID id, int tabIndex, TimeValue t);
};

class IKCtrlClassAccessor : public PBAccessor
{
public:
	void Get (PB2Value &v, ReferenceMaker *owner, ParamID id, int tabIndex, TimeValue t, Interval &valid);
	void Set(PB2Value &v, ReferenceMaker *owner, ParamID id, int tabIndex, TimeValue t);
};


// class to find all node dependents 
class FindNodesProc : public DependentEnumProc 
{
public:
	FindNodesProc(INodeTab* tab) {
		nodetab = tab;
	}
	int proc(ReferenceMaker *ref) {
			if (ref)
			{
			switch (ref->SuperClassID()) {
					case BASENODE_CLASS_ID:
						INode* n = (INode*)ref;
						nodetab->Append(1, &n, 5);
						return DEP_ENUM_SKIP;
			}
		}
		return DEP_ENUM_CONTINUE;
	}
private:
	INodeTab* nodetab;
};
